from onegov.pay.models.payment import Payment
from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Table
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import relationship


class Payable(object):
    """ Links the parent model with 0-n :class:`~onegov.pay.models.Payment`
    records through an automatically generated association table.

    """

    @declared_attr
    def payment(cls):
        name = 'payments_for_{}'.format(cls.__tablename__)
        key = '{}_id'.format(cls.__tablename__)
        target = '{}.id'.format(cls.__tablename__)
        backref = 'linked_{}'.format(cls.__tablename__)

        payment_association = Table(
            name,
            cls.metadata,
            Column(key, ForeignKey(target), primary_key=True, nullable=False),
            Column('payment_id', ForeignKey(Payment.id), nullable=False))

        Payment.register_link(backref, cls, payment_association, key)

        return relationship(
            argument=Payment,
            secondary=payment_association,
            backref=backref,
            cascade=False,
            uselist=False,
            passive_deletes=True)



from collections import namedtuple
from decimal import Decimal
from onegov.core.orm import Base
from onegov.core.orm.mixins import ContentMixin
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import UUID
from onegov.pay.utils import QueryChain
from sqlalchemy import Column
from sqlalchemy import Enum
from sqlalchemy import Numeric
from sqlalchemy import ForeignKey
from sqlalchemy import Text
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import joinedload
from sqlalchemy.orm import object_session
from uuid import uuid4


class RegisteredLink(namedtuple("RegisteredLink", ('cls', 'table', 'key'))):
    pass


class Payment(Base, TimestampMixin, ContentMixin):
    """ Represents a payment done through various means. """

    __tablename__ = 'payments'

    registered_links = {}

    #: the public id of the payment
    id = Column(UUID, primary_key=True, default=uuid4)

    #: the polymorphic source of the payment
    source = Column(Text, nullable=True)

    #: the amount to pay
    amount = Column(Numeric(precision=8, scale=2))

    #: the currency of the amount to pay
    currency = Column(Text, nullable=False, default='CHF')

    #: remote id of the payment
    remote_id = Column(Text, nullable=True)

    #: the state of the payment
    state = Column(
        Enum('open', 'paid', 'failed', 'cancelled', name='payment_state'),
        nullable=False,
        default='open'
    )

    #: the payment provider associated with the payment, if it is missing it
    #: means that the payment is out-of-band (say paid by cash)
    provider_id = Column(
        UUID, ForeignKey('payment_providers.id'), nullable=True
    )

    __mapper_args__ = {
        'polymorphic_on': source
    }

    @property
    def fee(self):
        """ The fee associated with this payment. The payment amount includes
        the fee. To get the net amount use the net_amount property.

        """
        return Decimal(0)

    @property
    def net_amount(self):
        return self.amount - self.fee

    @hybrid_property
    def paid(self):
        """ Our states are essentially one paid and n unpaid states (indicating
        various ways in which the payment can end up being unpaid).

        So this boolean acts as coarse filter to divide payemnts into the
        two states that really matter.

        """
        return self.state == 'paid'

    @staticmethod
    def register_link(link_name, linked_class, association_table, table_key):
        """ The :class:`~onegov.pay.models.payable.Payable` class registers
        all back-referenes through this method. This is useful for two reasons:

        1. We gain the ability to query all the linked records in one query.
           This is hard otherwise as each ``Payable`` class leads to its own
           association table which needs to be queried separately.

        2. We are able to reset all created backreferences. This is necessary
           during tests. SQLAlchemy keeps these references around and won't
           let us re-register the same model multiple times (which outside
           of tests is completely reasonable).

        Note that this is a staticmethod instead of a classmethod because
        we want all links to be managed through the actual ``Payment`` class,
        not through any of its subclasses.

        """
        Payment.registered_links[link_name] = RegisteredLink(
            cls=linked_class,
            table=association_table,
            key=table_key
        )

    @property
    def links(self):
        session = object_session(self)

        return QueryChain(tuple(
            session.query(link.cls)
            .filter_by(payment=self)
            .options(joinedload(link.cls.payment))
            for link in self.registered_links.values()
        ))

    @property
    def remote_url(self):
        """ Returns the url of this object on the payment provider. """

        raise NotImplementedError

    def sync(self, remote_obj=None):
        """ Updates the local payment information with the information from
        the remote payment provider.

        """

        raise NotImplementedError


class ManualPayment(Payment):
    """ A manual payment is a payment without associated payment provider.

    For example, a payment paid in cash.

    """
    __mapper_args__ = {'polymorphic_identity': 'manual'}

    def sync(self, remote_obj=None):
        pass



from onegov.pay.models.payable import Payable
from onegov.pay.models.payment import ManualPayment, Payment
from onegov.pay.models.payment_provider import PaymentProvider

__all__ = ('ManualPayment', 'Payable', 'Payment', 'PaymentProvider')



from onegov.core.collection import Pagination
from onegov.pay import Payment
from onegov.pay.utils import QueryChain
from sqlalchemy.orm import joinedload


class PayableCollection(Pagination):
    """ Provides a collection of all payable models. This collection is
    meant to be read-only, so there's no add/delete methods.

    To add payments to payable models just use the payment property and
    directly assign a new or an existing payment.

    """

    def __init__(self, session, cls='*', page=0):
        self.session = session
        self.cls = cls
        self.page = page

    @property
    def classes(self):
        if self.cls != '*':
            return (self.cls, )

        return tuple(link.cls for link in Payment.registered_links.values())

    def query(self):
        return QueryChain((
            self.session.query(cls).options(
                joinedload(cls.payment)
            )
            for cls in self.classes
        ))

    def __eq__(self, other):
        return self.cls == other.cls and self.page == other.page

    def subset(self):
        return self.query()

    @property
    def page_index(self):
        return self.page

    def page_by_index(self, index):
        return self.__class__(self.session, self.cls, index)



from onegov.core.orm import Base
from onegov.core.orm.mixins import ContentMixin
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.types import UUID
from onegov.pay.models.payment import Payment
from sqlalchemy import Boolean
from sqlalchemy import Column
from sqlalchemy import column
from sqlalchemy import Index
from sqlalchemy import Text
from sqlalchemy.orm import relationship
from uuid import uuid4


class PaymentProvider(Base, TimestampMixin, ContentMixin):
    """ Represents a payment provider. """

    __tablename__ = 'payment_providers'

    #: the public id of the payment provider
    id = Column(UUID, primary_key=True, default=uuid4)

    #: the polymorphic type of the provider
    type = Column(Text, nullable=True)

    #: true if this is the default provider (can only ever be one)
    default = Column(Boolean, nullable=False, default=False)

    __mapper_args__ = {
        'polymorphic_on': type
    }

    __table_args__ = (
        Index(
            'only_one_default_provider', 'default',
            unique=True, postgresql_where=column('default') == True
        ),
    )

    payments = relationship(
        'Payment',
        order_by='Payment.created',
        backref='provider',
        passive_deletes=True
    )

    @property
    def payment_class(self):
        assert type(self) is PaymentProvider, "Override this in subclasses"
        return Payment

    def payment(self, **kwargs):
        """ Creates a new payment using the correct model. """

        payment = self.payment_class(**kwargs)
        payment.provider = self

        return payment

    def adjust_price(self, price):
        """ Called by client implementations this method allows to adjust the
        price by adding a fee to it.

        By default no change is made.

        """

        return price

    def charge(self, amount, currency, token):
        """ Given a payment token, charges the customer and creates a payment
        which is returned.

        """
        raise NotImplementedError

    @property
    def title(self):
        """ The title of the payment provider (i.e. the product name). """
        raise NotImplementedError

    @property
    def url(self):
        """ The url to the backend of the payment provider. """
        raise NotImplementedError

    @property
    def public_identity(self):
        """ The public identifier of this payment provider. For example, the
        account name.

        """
        raise NotImplementedError

    @property
    def identity(self):
        """ Uniquely identifies this payment provider amongst other providers
        of the same type (say the private key of the api). Used to be able
        to tell if a new oauth connection is the same as an existing one.

        This identity is not meant to be displayed.

        """
        raise NotImplementedError

    @property
    def connected(self):
        """ Returns True if the provider is properly hooked up to the
        payment provider.

        """

    def checkout_button(self, label, amount, currency, action='submit',
                        **extra):
        """ Renders a checkout button which will store the token for the
        checkout as its own value if clicked.

        The action controls what happens after the token was successfully
        retrieved. The following actions are supported:

        - 'post': Submits the form surrounding the button.

        """
        raise NotImplementedError

    def prepare_oauth_request(self, redirect_url, success_url, error_url,
                              user_fields=None):
        """ Registers the oauth request with the oauth_gateway and returns
        an url that is ready to be used for the complete oauth request.

        """
        raise NotImplementedError

    def process_oauth_response(self, request_params):
        """ Processes the oauth response using the parameters passed by
        the returning oauth request via the gateway.

        """
        raise NotImplementedError



import logging
log = logging.getLogger('onegov.pay')  # noqa
log.addHandler(logging.NullHandler())  # noqa

from onegov.pay.errors import CARD_ERRORS
from onegov.pay.models import ManualPayment
from onegov.pay.models import Payable, Payment, PaymentProvider
from onegov.pay.collections import PaymentCollection, PayableCollection
from onegov.pay.collections import PaymentProviderCollection
from onegov.pay.integration import PayApp, process_payment
from onegov.pay.utils import Price


__all__ = (
    'log',
    'CARD_ERRORS',
    'ManualPayment',
    'Payable',
    'PayableCollection',
    'PayApp',
    'Payment',
    'PaymentCollection',
    'PaymentProvider',
    'PaymentProviderCollection',
    'Price',
    'process_payment'
)



from more.webassets import WebassetsApp
from onegov.core.orm import orm_cached
from onegov.pay import log
from onegov.pay import PaymentProvider
from onegov.pay.errors import CARD_ERRORS
from onegov.pay.models.payment import ManualPayment


class PayApp(WebassetsApp):
    """ Provides payment integration for
    :class:`onegov.core.framework.Framework` based applications.

    """

    def configure_payment_providers(self, **cfg):
        """ Configures the preconfigured parameters for payment providers.

        Takes one dictionary for each availble provider. Available providers
        can be found in the models/payment_providers folder. Additionally,
        it is possible to enable/disable custom payment providers for the
        whole site.

        For example::

            payment_providers_enabled: true
            payment_provider_defaults:
                stripe_connect:
                    client_id: foo
                    client_secret: bar

        Since multiple payment providers (even of the same type) may exist,
        and because some information stored on the payment providers need
        to be configured differently for each application_id (and possibly
        set up through OAuth) we only provide default parameters.

        When we create a new payment provider, these default values may be
        read by the payment provider.

        """

        self.payment_providers_enabled = cfg.get(
            'payment_providers_enabled', False)

        self.payment_provider_defaults = cfg.get(
            'payment_provider_defaults', {})

    @orm_cached(policy='on-table-change:payment_providers')
    def default_payment_provider(self):
        return self.session().query(PaymentProvider)\
            .filter_by(default=True).first()

    def adjust_price(self, price):
        """ Takes the given price object and adjusts it depending on the
        settings of the payment provider (for example, the fee might be
        charged to the user).

        """

        if self.default_payment_provider:
            return self.default_payment_provider.adjust_price(price)

        return price


@PayApp.webasset_path()
def get_js_path():
    return 'assets/js'


@PayApp.webasset('pay')
def get_pay_assets():
    yield 'stripe.js'


def process_payment(method, price, provider=None, token=None):
    """ Processes a payment using various methods, returning the processed
    payment or None.

    Available methods:

        'free': Payment may be done manually or by credit card
        'cc': Payment must be done by credit card
        'manual': Payment must be done manually

    """

    assert method in ('free', 'cc', 'manual') and price.amount > 0

    if method == 'free':
        method = token and 'cc' or 'manual'

    if method == 'manual':
        # manual payments do not carry a fee, so we ignore it
        return ManualPayment(amount=price.net_amount, currency=price.currency)

    if method == 'cc' and token:
        try:
            return provider.charge(
                amount=price.amount,
                currency=price.currency,
                token=token
            )
        except CARD_ERRORS:
            log.exception(
                "Processing {} through {} with token {} failed".format(
                    price,
                    provider.title,
                    token
                )
            )

    return None



from decimal import Decimal
from onegov.pay import Price


def test_price():
    assert Price(0, 'CHF') == Price(0, 'CHF')
    assert Price(0, 'CHF') < Price(1, 'CHF')
    assert Price(10, 'CHF') + Price(20, 'CHF') == Price(30, 'CHF')
    assert Price.zero() + Price(10, 'CHF') == Price(10, 'CHF')
    assert Price(10, 'CHF') - Price(20, 'CHF') == Price(-10, 'CHF')
    assert Price(10, 'CHF')[0] == Decimal(10)
    assert Price(10, 'CHF')[1] == 'CHF'

    amount, currency, fee = Price(10, 'CHF')
    assert amount == Decimal(10)
    assert currency == 'CHF'
    assert fee == Decimal(0)

    assert Price(10, 'CHF').as_dict() == {
        'amount': 10.0,
        'currency': 'CHF',
        'fee': 0
    }

    assert str(Price(10, 'CHF')) == '10.00 CHF'
    assert repr(Price(10, 'CHF')) == "Price(Decimal('10'), 'CHF')"



import logging

from onegov.pay import process_payment, Price
from onegov.pay import CARD_ERRORS


def test_process_manual_payment():
    payment = process_payment('manual', Price(10, 'CHF'))
    assert payment.source == 'manual'
    assert payment.amount == 10
    assert payment.currency == 'CHF'

    payment = process_payment('free', Price(10, 'CHF'))
    assert payment.source == 'manual'
    assert payment.amount == 10
    assert payment.currency == 'CHF'


def test_process_credit_card_payment_successfully():

    class Provider(object):
        def charge(self, amount, currency, token):
            return 'success'

    payment = process_payment('manual', Price(10, 'CHF'), Provider(), 'foobar')
    assert payment.source == 'manual'

    payment = process_payment('free', Price(10, 'CHF'), Provider(), 'foobar')
    assert payment == 'success'

    payment = process_payment('cc', Price(10, 'CHF'), Provider(), 'foobar')
    assert payment == 'success'


def test_process_credit_card_payment_error(capturelog):

    class Provider(object):
        title = 'Foobar'

        def charge(self, amount, currency, token):
            raise CARD_ERRORS[0](None, None, None)

    capturelog.setLevel(logging.ERROR, logger='onegov.pay')

    payment = process_payment('cc', Price(10, 'CHF'), Provider(), 'foobar')
    assert payment is None

    assert capturelog.records()[0].message\
        == 'Processing 10.00 CHF through Foobar with token foobar failed'



from onegov.pay.collections.payment import PaymentCollection
from onegov.pay.collections.payable import PayableCollection
from onegov.pay.collections.payment_provider import PaymentProviderCollection

__all__ = (
    'PayableCollection',
    'PaymentCollection',
    'PaymentProviderCollection'
)



from onegov.core.orm import Base
from onegov.core.orm import SessionManager
from onegov.core.orm.types import UUID
from onegov.pay import Payable, PayableCollection
from onegov.pay import PaymentCollection
from onegov.pay import PaymentProvider
from sqlalchemy import Column
from sqlalchemy import Text
from sqlalchemy.ext.declarative import declarative_base
from uuid import uuid4


def test_payment_collection_pagination(session):
    provider = PaymentProvider()

    session.add_all(
        provider.payment(amount=amount)
        for amount in range(100, 2000, 100)
    )

    payments = PaymentCollection(session)
    assert payments.query().count() == 19

    assert len(payments.batch) == 10
    assert len(payments.page_by_index(1).batch) == 9


def test_payment_collection_crud(session):
    payments = PaymentCollection(session)
    payment = payments.add(amount=100, provider=PaymentProvider())

    assert payment.amount == 100
    assert payments.query().count() == 1

    payments.delete(payment)
    assert payments.query().count() == 0


def test_payable_collection(postgres_dsn):

    MyBase = declarative_base()

    class Order(MyBase, Payable):
        __tablename__ = 'orders'

        id = Column(UUID, primary_key=True, default=uuid4)
        title = Column(Text)

    class Newspaper(MyBase, Payable):
        __tablename__ = 'newspapers'

        id = Column(UUID, primary_key=True, default=uuid4)
        title = Column(Text)

    mgr = SessionManager(postgres_dsn, Base)
    mgr.bases.append(MyBase)
    mgr.set_current_schema('foobar')
    session = mgr.session()

    provider = PaymentProvider()

    carbonara = Order(title="Carbonara")
    coca_cola = Order(title="Coca Cola")

    the_wapo = Newspaper(title="The Washington Post")
    ny_times = Newspaper(title="The New York Times")

    carbonara.payment = coca_cola.payment = provider.payment(amount=25)
    the_wapo.payment = provider.payment(amount=4.50)
    ny_times.payment = provider.payment(amount=5.50)

    session.add_all((carbonara, coca_cola, the_wapo, ny_times))
    session.flush()

    payables = PayableCollection(session)
    payables.batch_size = 3

    assert payables.query().count() == 4
    assert len(payables.batch) == 3

    payables = payables.page_by_index(1)
    payables.batch_size = 3
    assert len(payables.batch) == 1

    orders = PayableCollection(session, cls=Order)
    assert orders.query().count() == 2



from onegov.core.collection import GenericCollection
from onegov.pay.models import PaymentProvider


class PaymentProviderCollection(GenericCollection):
    """ Manages the payment providers. """

    @property
    def model_class(self):
        return PaymentProvider

    def as_default(self, provider):
        for other in self.query():
            other.default = False

        self.session.flush()
        provider.default = True

    def sync(self):
        """ Syncs all payments with the related payment providers.

        """

        for provider in self.query():
            provider.sync()



""" Contains upgrade tasks that are executed when the application is being
upgraded on the server. See :class:`onegov.core.upgrade.upgrade_task`.

"""

from onegov.core.upgrade import upgrade_task
from sqlalchemy import Column
from sqlalchemy import Text


@upgrade_task('Add remote_id field to payments')
def add_remote_id_field_to_payments(context):
    if not context.has_column('payments', 'remote_id'):
        context.operations.add_column('payments', Column(
            'remote_id', Text, nullable=True
        ))



import pytest
import transaction

from onegov.core.orm import Base, SessionManager
from onegov.core.orm.types import UUID
from onegov.pay.models import Payable, Payment, PaymentProvider, ManualPayment
from onegov.pay.collections import PaymentCollection
from sqlalchemy import Column
from sqlalchemy import Text
from sqlalchemy.ext.declarative import declarative_base
from uuid import uuid4
from sqlalchemy.exc import IntegrityError


def test_payment_with_different_bases(postgres_dsn):

    MyBase = declarative_base()

    class Order(MyBase, Payable):
        __tablename__ = 'orders'

        id = Column(UUID, primary_key=True, default=uuid4)
        title = Column(Text)

    class Subscription(Base, Payable):
        __tablename__ = 'subscriptions'

        id = Column(UUID, primary_key=True, default=uuid4)
        title = Column(Text)

    mgr = SessionManager(postgres_dsn, Base)
    mgr.bases.append(MyBase)
    mgr.set_current_schema('foobar')
    session = mgr.session()

    provider = PaymentProvider()

    apple = Order(title="Apple")
    pizza = Order(title="Pizza")
    kebab = Order(title="Kebab")
    times = Subscription(title="Times")

    apple.payment = provider.payment(amount=100)
    pizza.payment = provider.payment(amount=200)
    kebab.payment = apple.payment
    times.payment = pizza.payment

    session.add_all((apple, pizza, kebab, times))
    session.flush()

    assert session.query(Payment).count() == 2
    assert session.query(Order).count() == 3
    assert session.query(Subscription).count() == 1

    apple = session.query(Order).filter_by(title="Apple").one()
    pizza = session.query(Order).filter_by(title="Pizza").one()
    kebab = session.query(Order).filter_by(title="Kebab").one()
    times = session.query(Subscription).filter_by(title="Times").one()

    assert apple.payment.amount == 100
    assert pizza.payment.amount == 200
    assert kebab.payment.amount == 100
    assert times.payment.amount == 200


def test_payment_referential_integrity(postgres_dsn):

    MyBase = declarative_base()

    class Order(MyBase, Payable):
        __tablename__ = 'orders'

        id = Column(UUID, primary_key=True, default=uuid4)
        title = Column(Text)

    mgr = SessionManager(postgres_dsn, Base)
    mgr.bases.append(MyBase)
    mgr.set_current_schema('foobar')
    session = mgr.session()

    apple = Order(title="Apple", payment=PaymentProvider().payment(amount=100))
    session.add(apple)
    transaction.commit()

    with pytest.raises(IntegrityError):
        session.delete(session.query(PaymentProvider).one())
        session.flush()

    transaction.abort()

    # as a precaution we only allow deletion of elements after the payment
    # has been explicitly deleted
    with pytest.raises(IntegrityError):
        session.delete(session.query(Order).one())
        session.flush()

    transaction.abort()

    with pytest.raises(IntegrityError):
        session.delete(session.query(Order).one())
        session.delete(session.query(Payment).one())
        session.flush()

    transaction.abort()

    session.delete(session.query(Payment).one())
    session.delete(session.query(Order).one())
    transaction.commit()

    assert not list(session.execute("select * from orders"))
    assert not list(session.execute("select * from payments"))
    assert not list(session.execute("select * from payments_for_orders"))

    mgr.dispose()


def test_backref(postgres_dsn):

    MyBase = declarative_base()

    class Product(MyBase, Payable):
        __tablename__ = 'products'

        id = Column(UUID, primary_key=True, default=uuid4)
        title = Column(Text)

    class Part(MyBase, Payable):
        __tablename__ = 'parts'

        id = Column(UUID, primary_key=True, default=uuid4)
        title = Column(Text)

    mgr = SessionManager(postgres_dsn, Base)
    mgr.bases.append(MyBase)
    mgr.set_current_schema('foobar')
    session = mgr.session()

    provider = PaymentProvider()

    car = Product(title="Car", payment=provider.payment(amount=10000))
    nut = Part(title="Nut", payment=provider.payment(amount=10))
    session.add_all((car, nut))
    session.flush()

    payments = session.query(Payment).all()
    assert [t.title for p in payments for t in p.linked_products] == ["Car"]
    assert [t.title for p in payments for t in p.linked_parts] == ["Nut"]

    assert len(car.payment.linked_products) == 1
    assert len(car.payment.linked_parts) == 0
    assert len(nut.payment.linked_products) == 0
    assert len(nut.payment.linked_parts) == 1

    assert car.payment.links.count() == 1
    assert car.payment.links.first().title == "Car"
    assert nut.payment.links.count() == 1
    assert nut.payment.links.first().title == "Nut"

    assert len(PaymentCollection(session).payment_links_by_batch()) == 2

    session.delete(nut.payment)
    nut.payment = car.payment
    session.flush()

    assert len(car.payment.linked_products) == 1
    assert len(car.payment.linked_parts) == 1
    assert len(nut.payment.linked_products) == 1
    assert len(nut.payment.linked_parts) == 1

    assert car.payment.links.count() == 2
    assert {r.title for r in car.payment.links} == {"Car", "Nut"}

    assert len(PaymentCollection(session).payment_links_by_batch()) == 1

    mgr.dispose()


def test_manual_payment(postgres_dsn):

    MyBase = declarative_base()

    class Product(MyBase, Payable):
        __tablename__ = 'products'

        id = Column(UUID, primary_key=True, default=uuid4)
        title = Column(Text)

    mgr = SessionManager(postgres_dsn, Base)
    mgr.bases.append(MyBase)
    mgr.set_current_schema('foobar')
    session = mgr.session()

    car = Product(title="Car", payment=ManualPayment(amount=10000))
    session.add(car)
    session.flush()

    payments = session.query(Payment).all()
    assert [t.title for p in payments for t in p.linked_products] == ["Car"]



from collections import namedtuple
from decimal import Decimal
from functools import total_ordering
from sqlalchemy_utils import QueryChain as QueryChainBase


class QueryChain(QueryChainBase):
    """ Extends SQLAlchemy Utils' QueryChain with some extra methods. """

    def slice(self, start, end):
        return self[start:end]

    def first(self):
        return next((o for o in self), None)

    def all(self):
        return tuple(self)


@total_ordering
class Price(namedtuple('PriceBase', ('amount', 'currency', 'fee'))):
    """ A single price.

    The amount includes the fee. To get the net amount use the net_amount
    property.

    """

    def __new__(cls, amount, currency, fee=0):
        return super().__new__(cls, Decimal(amount), currency, Decimal(fee))

    def __bool__(self):
        return self.amount and True or False

    def __lt__(self, other):
        return self.amount < other.amount

    def __add__(self, other):
        assert self.currency is None or self.currency == other.currency
        return self.__class__(
            self.amount + other.amount,
            self.currency or other.currency
        )

    def __sub__(self, other):
        assert self.currency == other.currency
        return self.__class__(self.amount - other.amount, self.currency)

    def __str__(self):
        return '{:.2f} {}'.format(self.amount, self.currency)

    def __repr__(self):
        return 'Price({}, {})'.format(repr(self.amount), repr(self.currency))

    @classmethod
    def zero(cls):
        return cls(0, None)

    def as_dict(self):
        return {
            'amount': float(self.amount),
            'currency': self.currency,
            'fee': float(self.fee)
        }

    @property
    def net_amount(self):
        return self.amount - self.fee



import logging
import pytest
import requests_mock
import transaction

from onegov.pay.models.payment_providers.stripe import (
    StripeConnect,
    StripeFeePolicy,
    StripeCaptureManager
)
from purl import URL
from unittest import mock
from urllib.parse import quote


def test_oauth_url():
    provider = StripeConnect(client_id='foo', client_secret='bar')

    url = provider.oauth_url('https://handle-incoming-request')
    assert 'response_type=code' in url
    assert 'handle-incoming-request' in url
    assert 'scope=read_write' in url
    assert 'client_id=foo' in url
    assert 'client_secret=bar' in url

    url = provider.oauth_url('https://foo', 'bar', {'email': 'foo@bar.org'})
    assert 'state=bar' in url
    assert 'stripe_user%5Bemail%5D=foo%40bar.org' in url


def test_process_oauth_response():
    provider = StripeConnect(
        client_id='foo',
        client_secret='bar',
        oauth_gateway='https://oauth.onegovcloud.ch/',
        oauth_gateway_secret='foo',
    )

    with pytest.raises(RuntimeError) as e:
        provider.process_oauth_response(
            {
                'error': 'foo',
                'error_description': 'bar',
            }
        )

    assert e.value.args == ('Stripe OAuth request failed (foo: bar)', )

    with mock.patch('stripe.OAuth.token', return_value={
        'scope': 'read_write',
        'stripe_publishable_key': 'pubkey',
        'stripe_user_id': 'uid',
        'refresh_token': 'rtoken',
        'access_token': 'atoken',
    }):
        provider.process_oauth_response({
            'code': '0xdeadbeef',
            'oauth_redirect_secret': 'foo'
        })

        assert provider.publishable_key == 'pubkey'
        assert provider.user_id == 'uid'
        assert provider.refresh_token == 'rtoken'
        assert provider.access_token == 'atoken'


def test_stripe_fee_policy():
    assert StripeFeePolicy.from_amount(100) == 3.2
    assert StripeFeePolicy.compensate(100) == 103.3
    assert StripeFeePolicy.compensate(33.33) == 34.63


def test_stripe_capture_good_charge():
    class GoodCharge(object):
        captured = False

        def capture(self):
            self.captured = True

    charge = GoodCharge()

    with mock.patch('stripe.Charge.retrieve', return_value=charge):
        StripeCaptureManager.capture_charge('foo', 'bar')
        assert not charge.captured

        transaction.commit()
        assert charge.captured


def test_stripe_capture_evil_charge(capturelog):
    capturelog.setLevel(logging.ERROR, logger='onegov.pay')

    class EvilCharge(object):
        def capture(self):
            assert False

    charge = EvilCharge()

    with mock.patch('stripe.Charge.retrieve', return_value=charge):
        StripeCaptureManager.capture_charge('foo', 'bar')

        transaction.commit()
        assert capturelog.records()[2].message\
            == 'Stripe charge with capture id bar failed'


def test_stripe_capture_negative_vote():
    with mock.patch('stripe.Charge.retrieve', side_effect=AssertionError()):
        StripeCaptureManager.capture_charge('foo', 'bar')

        with pytest.raises(AssertionError):
            transaction.commit()


def test_stripe_prepare_oauth_request():
    stripe = StripeConnect()
    stripe.oauth_gateway = 'https://oauth.example.org'
    stripe.oauth_gateway_auth = 'gateway_auth'
    stripe.oauth_gateway_secret = 'gateway_secret'
    stripe.client_id = 'client_id'
    stripe.client_secret = 'client_secret'

    with requests_mock.Mocker() as m:
        m.post('https://oauth.example.org/register/gateway_auth', json={
            'token': '0xdeadbeef'
        })

        url = stripe.prepare_oauth_request(
            redirect_uri='https://endpoint',
            success_url='https://success',
            error_url='https://error'
        )

        assert quote('https://oauth.example.org/redirect', safe='') in url

        url = URL(url)
        assert url.query_param('state') == '0xdeadbeef'
        assert url.query_param('scope') == 'read_write'
        assert url.query_param('client_id') == 'client_id'



from collections import defaultdict
from onegov.core.collection import GenericCollection, Pagination
from onegov.pay.models import Payment
from sqlalchemy import desc
from sqlalchemy.orm import joinedload
from sqlalchemy.orm import undefer


class PaymentCollection(GenericCollection, Pagination):
    """ Manages the payment records.

    To render a list of payments you might want to also consider the
    :class:`onegov.pay.collection.payable.Paybale` collection, which renders
    payments by loading the linked records first.

    """

    def __init__(self, session, source='*', page=0):
        super().__init__(session)
        self.source = source
        self.page = page

    @property
    def model_class(self):
        return Payment.get_polymorphic_class(self.source, Payment)

    def add(self, **kwargs):
        if self.source != '*':
            kwargs.setdefault('source', self.source)
        return super().add(**kwargs)

    def __eq__(self, other):
        return self.source == other.source and self.page == other.page

    def subset(self):
        q = self.query().order_by(desc(Payment.created))
        q = q.options(joinedload(Payment.provider))
        q = q.options(undefer(Payment.created))
        return q

    @property
    def page_index(self, page):
        return self.page

    def page_by_index(self, index):
        return self.__class__(self.session, self.source, index)

    def payment_links_by_batch(self, batch=None):
        """ A more efficient way of loading all links of the given batch
        (compared to loading payment.links one by one).

        """

        batch = batch or self.batch

        if not batch:
            return None

        payment_links = defaultdict(list)

        for link in Payment.registered_links.values():
            targets = self.session.query(
                getattr(link.table.columns, link.key)
            ).filter(
                link.table.columns.payment_id.in_(tuple(
                    p.id for p in self.batch
                ))
            )

            q = self.session.query(link.cls)
            q = q.filter(link.cls.id.in_(targets.subquery()))
            q = q.options(joinedload(link.cls.payment))

            for link in q:
                payment_links[link.payment.id].append(link)

        return payment_links



import requests
import stripe
import transaction

from cached_property import cached_property
from contextlib import contextmanager
from datetime import datetime
from decimal import Decimal
from html import escape
from onegov.core.orm.mixins import meta_property
from onegov.pay import log
from onegov.pay.models.payment import Payment
from onegov.pay.models.payment_provider import PaymentProvider
from onegov.pay.utils import Price
from sqlalchemy.orm import object_session
from uuid import UUID, uuid4, uuid5


@contextmanager
def stripe_api_key(key):
    old_key = stripe.api_key
    stripe.api_key = key
    yield
    stripe.api_key = old_key


# instantiate once to get keep-alive support
stripe.default_http_client = stripe.http_client.RequestsClient()


# our stripe payment ids are generated by using the token and a UUID namespace
STRIPE_NAMESPACE = UUID('aebb1a32-282b-4521-838d-92a1136624d1')


class StripeCaptureManager(object):
    """ Captures an open stripe charge when the transaction finishes.

    If there is an error during this step, it is logged, but the transaction
    still continues successfully.

    The user is then supposed to manually capture the charge.

    """

    transaction_manager = transaction.manager

    def __init__(self, api_key, charge_id):
        self.api_key = api_key
        self.charge_id = charge_id

    @classmethod
    def capture_charge(cls, api_key, charge_id):
        transaction.get().join(cls(api_key, charge_id))

    def sortKey(self):
        return 'charge'

    def tpc_vote(self, transaction):
        with stripe_api_key(self.api_key):
            self.charge = stripe.Charge.retrieve(self.charge_id)

    def tpc_finish(self, transaction):
        try:
            with stripe_api_key(self.api_key):
                self.charge.capture()
        except:
            # we can never fail or we might end up with an incosistent
            # database -> so must swallow any errors and report them
            log.exception("Stripe charge with capture id {} failed".format(
                self.charge_id
            ))

    def commit(self, transaction):
        pass

    def abort(self, transaction):
        pass

    def tpc_begin(self, transaction):
        pass

    def tpc_abort(self, transaction):
        pass


class StripeFeePolicy(object):
    """ All stripe fee calculations in one place (should they ever change). """

    percentage = 0.029
    fixed = 0.3

    @classmethod
    def from_amount(cls, amount):
        """ Gets the fee for the given amount. """

        return round(float(amount) * cls.percentage + cls.fixed, 2)

    @classmethod
    def compensate(cls, amount):
        """ Increases the amount in such a way that the stripe fee is included
        in the effective charge (that is, the user paying the charge is paying
        the fee as well).

        """

        return round((float(amount) + cls.fixed) / (1 - cls.percentage), 2)


class StripePayment(Payment):
    __mapper_args__ = {'polymorphic_identity': 'stripe_connect'}

    fee_policy = StripeFeePolicy

    #: the date of the payout
    payout_date = meta_property('payout_date')

    #: the id of the payout
    payout_id = meta_property('payout_id')

    #: the fee deducted by stripe
    effective_fee = meta_property('effective_fee')

    @property
    def fee(self):
        """ The calculated fee or the effective fee if available.

        The effective fee is taken from the payout records. In practice
        these values should always be the same.

        """

        if self.effective_fee:
            return Decimal(self.effective_fee)

        return Decimal(self.fee_policy.from_amount(self.amount))

    @property
    def remote_url(self):
        if self.provider.livemode:
            base = 'https://dashboard.stripe.com/payments/{}'
        else:
            base = 'https://dashboard.stripe.com/test/payments/{}'

        return base.format(self.remote_id)

    @property
    def charge(self):
        with stripe_api_key(self.provider.access_token):
            return stripe.Charge.retrieve(self.remote_id)

    def refund(self):
        with stripe_api_key(self.provider.access_token):
            refund = stripe.Refund.create(charge=self.remote_id)
            self.state = 'cancelled'
            return refund

    def sync(self, remote_obj=None):
        charge = remote_obj or self.charge

        if not charge.captured:
            self.state = 'open'

        elif charge.refunded:
            self.state = 'cancelled'

        elif charge.status == 'failed':
            self.state = 'failed'

        elif charge.captured and charge.paid:
            self.state = 'paid'


class StripeConnect(PaymentProvider):

    __mapper_args__ = {'polymorphic_identity': 'stripe_connect'}

    fee_policy = StripeFeePolicy

    #: The Stripe Connect client id
    client_id = meta_property('client_id')

    #: The API key of the connect user
    client_secret = meta_property('client_secret')

    #: The oauth_redirect gateway in use (see seantis/oauth_redirect on github)
    oauth_gateway = meta_property('oauth_gateway')

    #: The auth code required by oauth_redirect
    oauth_gateway_auth = meta_property('oauth_gateway_auth')

    #: The oauth_redirect secret that should be used
    oauth_gateway_secret = meta_property('oauth_gateway_secret')

    #: The authorization code provided by OAuth
    authorization_code = meta_property('authorization_code')

    #: The public stripe key
    publishable_key = meta_property('publishable_key')

    #: The stripe user id as confirmed by OAuth
    user_id = meta_property('user_id')

    #: The refresh token provided by OAuth
    refresh_token = meta_property('refresh_token')

    #: The access token provieded by OAuth
    access_token = meta_property('access_token')

    #: The id of the latest processed balance transaction
    latest_payout = meta_property('latest_payout')

    #: Should the fee be charged to the customer or not?
    charge_fee_to_customer = meta_property('charge_fee_to_customer')

    def adjust_price(self, price):
        if self.charge_fee_to_customer:
            new_price = self.fee_policy.compensate(price.amount)
            new_fee = self.fee_policy.from_amount(new_price)

            return Price(new_price, price.currency, new_fee)

        return price

    @property
    def livemode(self):
        return not self.access_token.startswith('sk_test')

    @property
    def payment_class(self):
        return StripePayment

    @property
    def title(self):
        return 'Stripe Connect'

    @property
    def url(self):
        return 'https://dashboard.stripe.com/'

    @property
    def public_identity(self):
        account = self.account
        return ' / '.join((account.business_name, account.email))

    @property
    def identity(self):
        return self.user_id

    @cached_property
    def account(self):
        with stripe_api_key(self.access_token):
            return stripe.Account.retrieve(id=self.user_id)

    @property
    def connected(self):
        return self.account and True or False

    def charge(self, amount, currency, token):
        session = object_session(self)

        payment = self.payment(
            id=uuid5(STRIPE_NAMESPACE, token),
            amount=amount,
            currency=currency,
            state='open'
        )

        with stripe_api_key(self.access_token):
            charge = stripe.Charge.create(
                amount=round(amount * 100, 0),
                currency=currency,
                source=token,
                capture=False,
                idempotency_key=token,
                metadata={
                    'payment_id': payment.id.hex
                }
            )

        StripeCaptureManager.capture_charge(self.access_token, charge.id)
        payment.remote_id = charge.id

        # we do *not* want to lose this information, so even though the
        # caller should make sure the payment is stored, we make sure
        session.add(payment)

        return payment

    def checkout_button(self, label, amount, currency, action='submit',
                        **extra):
        """ Generates the html for the checkout button. """

        extra['amount'] = round(amount * 100, 0)
        extra['currency'] = currency
        extra['key'] = self.publishable_key

        attrs = {
            'data-stripe-{}'.format(key): str(value)
            for key, value in extra.items()
        }
        attrs['data-action'] = action

        return """
            <input type="hidden" name="payment_token" id="{target}">
            <button class="checkout-button stripe-connect"
                    data-target-id="{target}"
                    {attrs}>{label}</button>
        """.format(
            label=escape(label),
            attrs=' '.join(
                '{}="{}"'.format(escape(k), escape(v))
                for k, v in attrs.items()
            ),
            target=uuid4().hex
        )

    def oauth_url(self, redirect_uri, state=None, user_fields=None):
        """ Generates an oauth url to be shown in the browser. """

        return stripe.OAuth.authorize_url(
            client_id=self.client_id,
            client_secret=self.client_secret,
            scope='read_write',
            redirect_uri=redirect_uri,
            stripe_user=user_fields,
            state=state
        )

    def prepare_oauth_request(self, redirect_uri, success_url, error_url,
                              user_fields=None):
        """ Registers the oauth request with the oauth_gateway and returns
        an url that is ready to be used for the complete oauth request.

        """
        register = '{}/register/{}'.format(
            self.oauth_gateway,
            self.oauth_gateway_auth)

        assert self.oauth_gateway \
            and self.oauth_gateway_auth \
            and self.oauth_gateway_secret

        payload = {
            'url': redirect_uri,
            'secret': self.oauth_gateway_secret,
            'method': 'GET',
            'success_url': success_url,
            'error_url': error_url
        }

        response = requests.post(register, json=payload)
        assert response.status_code == 200

        return self.oauth_url(
            redirect_uri='{}/redirect'.format(self.oauth_gateway),
            state=response.json()['token'],
            user_fields=user_fields
        )

    def process_oauth_response(self, request_params):
        """ Takes the parameters of an incoming oauth request and stores
        them on the payment provider if successful.

        """

        if 'error' in request_params:
            raise RuntimeError("Stripe OAuth request failed ({}: {})".format(
                request_params['error'], request_params['error_description']
            ))

        assert request_params['oauth_redirect_secret'] \
            == self.oauth_gateway_secret

        self.authorization_code = request_params['code']

        with stripe_api_key(self.client_secret):
            token = stripe.OAuth.token(
                grant_type='authorization_code',
                code=self.authorization_code,
            )

        assert token['scope'] == 'read_write'

        self.publishable_key = token['stripe_publishable_key']
        self.user_id = token['stripe_user_id']
        self.refresh_token = token['refresh_token']
        self.access_token = token['access_token']

    def sync(self):
        session = object_session(self)
        self.sync_payment_states(session)
        self.sync_payouts(session)

    def sync_payment_states(self, session):

        def payments(ids):
            q = session.query(self.payment_class)
            q = q.filter(self.payment_class.id.in_(ids))

            return q

        charges = self.paged(
            stripe.Charge.list,
            limit=50,
            include=lambda r: 'payment_id' in r.metadata
        )

        by_payment = {}

        for charge in charges:
            by_payment[charge.metadata['payment_id']] = charge

        for payment in payments(by_payment.keys()):
            payment.sync(remote_obj=by_payment[payment.id.hex])

    def sync_payouts(self, session):

        payouts = self.paged(stripe.Payout.list, limit=50, status='paid')
        latest_payout = None

        paid_charges = {}

        for payout in payouts:
            if latest_payout is None:
                latest_payout = payout

            if payout.id == self.latest_payout:
                break

            transactions = self.paged(
                stripe.BalanceTransaction.list,
                limit=50,
                payout=payout.id,
                type='charge'
            )

            for charge in transactions:
                paid_charges[charge.source] = (
                    datetime.fromtimestamp(payout.arrival_date),
                    payout.id,
                    charge.fee / 100
                )

        if paid_charges:
            q = session.query(self.payment_class)
            q = q.filter(self.payment_class.remote_id.in_(paid_charges.keys()))

            for p in q:
                p.payout_date, p.payout_id, p.fee = paid_charges[p.remote_id]

        self.latest_payout = latest_payout and latest_payout.id

    def paged(self, method, include=lambda record: True, **kwargs):
        with stripe_api_key(self.access_token):
            records = method(**kwargs)
            records = (r for r in records.auto_paging_iter())
            records = (r for r in records if include(r))

            yield from records



import pytest

from onegov.pay.models import Payment


@pytest.fixture(scope='function', autouse=True)
def reset_payment():
    yield

    # during testing we need to reset the links created on the payment
    # model - in reality this is not an issue as we don't define the same
    # models over and over
    classes = [Payment]

    while classes:
        cls = classes.pop()

        for key in Payment.registered_links:
            delattr(cls, key)
            del cls.__mapper__._props[key]

        classes.extend(cls.__subclasses__())

    Payment.registered_links.clear()
